#include once "../../Include/VM/VmProcess.bi"
#include once "../../Include/VM/VmRuntime.bi"



dim shared OpCodeHandlers as sub (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr) ptr
dim shared Processes as LinkedNode ptr
sub VmRuntime_Init()
	Processes=0
	ConsoleWrite(@"Initializing runtime ... ")    

	OpCodeHandlers = MEM_ALLOC(sizeof(any ptr)*255)
	
	OpCodeHandlers[OPCode._NOP]=@Handle_NOP
	OpCodeHandlers[OPCode._MOV]=@Handle_MOV
	OpCodeHandlers[OPCode._CLE]=@Handle_CLE
    OpCodeHandlers[OPCode._CLA]=@Handle_CLA
    OpCodeHandlers[OPCode._CLB]=@Handle_CLB
    OpCodeHandlers[OPCode._CLZ]=@Handle_CLZ
    OpCodeHandlers[OPCode._SLE]=@Handle_SLE
    OpCodeHandlers[OPCode._SLA]=@Handle_SLA
    OpCodeHandlers[OPCode._SLB]=@Handle_SLB
    OpCodeHandlers[OPCode._SLZ]=@Handle_SLZ
	OpCodeHandlers[OPCode._ITOS]=@Handle_ITOS
	OpCodeHandlers[OPCode._STOI]=@Handle_STOI
    OpCodeHandlers[OPCode._ADD]=@Handle_ADD
    OpCodeHandlers[OPCode._FADD]=@Handle_FADD
    OpCodeHandlers[OPCode._SUB]=@Handle_SUB
    OpCodeHandlers[OPCode._FSUB]=@Handle_FSUB
    OpCodeHandlers[OPCode._MUL]=@Handle_MUL
    OpCodeHandlers[OPCode._FMUL]=@Handle_FMUL
	OpCodeHandlers[OPCode._DIV]=@Handle_DIV
    OpCodeHandlers[OPCode._FDIV]=@Handle_FDIV
	OpCodeHandlers[OPCode._INC]=@Handle_INC
    OpCodeHandlers[OPCode._FINC]=@Handle_FINC
    OpCodeHandlers[OPCode._DEC]=@Handle_DEC
    OpCodeHandlers[OPCode._FDEC]=@Handle_FDEC
    OpCodeHandlers[OPCode._LOOP]=@Handle_LOOP
    OpCodeHandlers[OPCode._JMP]=@Handle_JMP
    OpCodeHandlers[OPCode._JE]=@Handle_JE
    OpCodeHandlers[OPCode._JNE]=@Handle_JNE
    OpCodeHandlers[OPCode._JA]=@Handle_JA
    OpCodeHandlers[OPCode._JNA]=@Handle_JNA
    OpCodeHandlers[OPCode._JB]=@Handle_JNAE
    OpCodeHandlers[OPCode._JNB]=@Handle_JAE
    OpCodeHandlers[OPCode._JAE]=@Handle_JAE
    OpCodeHandlers[OPCode._JNAE]=@Handle_JNAE
    OpCodeHandlers[OPCode._JBE]=@Handle_JNA
    OpCodeHandlers[OPCode._JNBE]=@Handle_JA
    OpCodeHandlers[OPCode._AND]=@Handle_AND
    OpCodeHandlers[OPCode._OR]=@Handle_OR
    OpCodeHandlers[OPCode._XOR]=@Handle_XOR
    OpCodeHandlers[OPCode._NEG]=@Handle_NEG
    OpCodeHandlers[OPCode._NOT]=@Handle_NOT
    OpCodeHandlers[OPCode._SHR]=@Handle_SHR
    OpCodeHandlers[OPCode._SHL]=@Handle_SHL
    OpCodeHandlers[OPCode._PUSH]=@Handle_PUSH
    OpCodeHandlers[OPCode._POP]=@Handle_POP
    OpCodeHandlers[OPCode._PUSHF]=@Handle_PUSHF
    OpCodeHandlers[OPCode._POPF]=@Handle_POPF
    OpCodeHandlers[OPCode._PUSHA]=@Handle_PUSHA
    OpCodeHandlers[OPCode._POPA]=@Handle_POPA
    OpCodeHandlers[OPCode._XCALL]=@Handle_XCALL
    OpCodeHandlers[OPCode._CALL]=@Handle_CALL
	OpCodeHandlers[OPCode._RET]=@Handle_RET
    OpCodeHandlers[OPCode._CMP]=@Handle_CMP    
    OpCodeHandlers[OPCode._INT]=@Handle_INT    
    OpCodeHandlers[OPCode._HLT]=@Handle_HLT    
    OpCodeHandlers[OPCode._LMOD]=@Handle_LMOD    
        
	Processes = LinkedList_AddItem(Processes,VMProcess_Create(0))
	
    ConsoleWriteLine(@"DONE")    
end sub

sub VmRunTime_MainLoop()
	dim processNode as LinkedNode ptr
	dim threadNode as LinkedNode ptr
	dim process as VmProcess ptr
	dim thread as VmThread ptr
	processNode = Processes
	
	ConsoleWrite(@"First process at 0x")
	ConsoleWriteNumber(cast(unsigned integer, Processes),16)
	ConsoleNewLine()
	
	
	while(processNode<>0)
		process = processNode->Item
		threadNode=process->Threads
		while(threadNode<>0)
			thread=threadNode->Item
			VmRunTime_ForeachThread(thread)
			threadNode=threadNode->NextItem
		wend
		processNode=processNode->NextItem
	wend
end sub

sub VmRunTime_ForeachThread(theThread as VmThread ptr)
	
	dim instruction as VmInstructionBlock
	dim workingMemory as byte ptr
	workingMemory=0
	
	if (theThread->ThreadRegs.SEG=0) then
		instruction=theThread->MainProcess->Instructions[theThread->ThreadRegs.EIP / sizeof(VmInstructionBlock)]
		workingMemory = theThread->MainProcess->MemoryData
	else
		'on prend le module qui correspond à seg
		'c'est linstruction à cette adresse (de eip de mon thread)
		'c'est la mémoire de mon module
	end if
	theThread->ThreadRegs.EIP = theThread->ThreadRegs.EIP + sizeof(VmInstructionBlock)
	VmRuntime_Run(theThread->ThreadHandle,instruction,theThread->ThreadRegs,workingMemory,theThread->ThreadStack)
end sub

sub VmRuntime_Run(threadHandle as unsigned integer,instruction as VmInstructionBlock,regs as RegistersMemory,appMem as byte ptr, stackMem as StackMemory ptr)
	if OpCodeHandlers[instruction.OPCODE] <> 0 then
		ConsoleWrite(@"Runing from threadHandle 0x")
		ConsoleWriteNumber(threadHandle,16)
		ConsoleNewLine()
	else
		ConsoleWriteLine(@"Invalid opcode")
	end if
end sub

sub Handle_NOP (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_MOV (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_CLE (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_CLA (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_CLB (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_CLZ (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_SLE (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_SLA (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_SLB (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_SLZ (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_ITOS (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_STOI (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_ADD (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_FADD (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_SUB (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_FSUB (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_MUL (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_FMUL (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_DIV (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_FDIV (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_INC (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_FINC (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_DEC (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_FDEC (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_LOOP (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_JMP (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_JE (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_JNE (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_JA (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_JNA (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_JAE (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_JNAE (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_AND (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_OR (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_XOR (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_NEG (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_NOT (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_SHL (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_SHR (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_PUSH (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_POP (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_PUSHF (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_POPF (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub      

sub Handle_PUSHA (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_POPA (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_CALL (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub      

sub Handle_XCALL (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_RET (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_CMP (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_INT (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_HLT (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub

sub Handle_LMOD (threadHandle as unsigned integer,instruction as VmInstructionBlock, regs as RegistersMemory, appMem as byte ptr, stackMem as StackMemory ptr)
end sub